/*
 * Handler routines.
 *
 * Copyright (C) 2015 Nicholas FitzRoy-Dale <wzdd.code@lardcave.net>
 * Copyright (C) 2015, 2016 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

.text
.extern irq_handle
.extern current_registers
.extern current_stack_pointer
.extern current_task
.globl interrupt_handler
.set noreorder
.set noat

interrupt_handler:
	/* gp should have been set in the entrypoint. */

	jal save_state
	nop

	/* Invoke the handler. */

	jal irq_handle
	nop

	/* Switch the stack pointer. */

	la $k0, current_stack_pointer
	lw $k1, 0($k0)					/* &stack_pointers[current_task] */
	lw $sp, 0($k1)

	/* Set the current task ASID. */

	la $k0, current_task
	lw $k1, 0($k0)
	mtc0 $k1, $10					/* CP0_ENTRYHI */
	nop

	/* Obtain the current task's registers. */

	j load_and_return
	nop

save_state:
	/* Obtain a store of registers for the current task. */

	la $k0, current_registers
	lw $k1, 0($k0)

	sw $at, 4($k1)
	sw $v0, 8($k1)
	sw $v1, 12($k1)
	sw $a0, 16($k1)
	sw $a1, 20($k1)
	sw $a2, 24($k1)
	sw $a3, 28($k1)
	sw $t0, 32($k1)
	sw $t1, 36($k1)
	sw $t2, 40($k1)
	sw $t3, 44($k1)
	sw $t4, 48($k1)
	sw $t5, 52($k1)
	sw $t6, 56($k1)
	sw $t7, 60($k1)
	sw $s0, 64($k1)
	sw $s1, 68($k1)
	sw $s2, 72($k1)
	sw $s3, 76($k1)
	sw $s4, 80($k1)
	sw $s5, 84($k1)
	sw $s6, 88($k1)
	sw $s7, 92($k1)
	sw $t8, 96($k1)

	/* t9, gp and ra are already saved */

	sw $fp, 108($k1)

	mfc0 $k0, $14					/* CP0_EPC */
	nop
	sw $k0, 116($k1)

	mflo $k0
	sw $k0, 120($k1)
	mfhi $k0
	sw $k0, 124($k1)

	j $ra
	nop

load_and_return:
	/* Obtain a store of registers for the current task. */

	la $k0, current_registers
	lw $k1, 0($k0)

	lw $at, 4($k1)
	lw $v0, 8($k1)
	lw $v1, 12($k1)
	lw $a0, 16($k1)
	lw $a1, 20($k1)
	lw $a2, 24($k1)
	lw $a3, 28($k1)
	lw $t0, 32($k1)
	lw $t1, 36($k1)
	lw $t2, 40($k1)
	lw $t3, 44($k1)
	lw $t4, 48($k1)
	lw $t5, 52($k1)
	lw $t6, 56($k1)
	lw $t7, 60($k1)
	lw $s0, 64($k1)
	lw $s1, 68($k1)
	lw $s2, 72($k1)
	lw $s3, 76($k1)
	lw $s4, 80($k1)
	lw $s5, 84($k1)
	lw $s6, 88($k1)
	lw $s7, 92($k1)
	lw $t8, 96($k1)
	lw $t9, 100($k1)
	lw $gp, 104($k1)
	lw $fp, 108($k1)
	lw $ra, 112($k1)

	lw $k0, 116($k1)
	mtc0 $k0, $14					/* CP0_EPC */
	nop

	sw $k0, 120($k1)
	mtlo $k0
	sw $k0, 124($k1)
	mthi $k0

	eret
	nop

.set reorder
.set at
